/*****************************************************************************
Copyright 1987, 1988, 1989, 1990, 1991 by Digital Equipment Corp., Maynard, MA

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

*****************************************************************************/
/*
 *  VERSION:    V1.0
 *
 *  FACILITY:   priv_utils
 *
 *  ABSTRACT:
 *
 *      This module contains utility routines to get a privilege mask 
 *      and raise or lower specific privileges.
 *
 *
 *  AUTHOR:     Alan Jamison
 *  DATE:       12-Mar-1990
 *
 *  MODIFIED BY:
 *
 *      Kenneth B. Miller 26-May-1990 "Portable-ized it for XTrap Library"
 *
 */

/*---------------------*
 *  Include Files      *
 *---------------------*/

#include "stdio.h"
#ifdef VMS /* to end-of-file */
#include "stdarg.h"     /* Standard definitions for variable argument lists */
#include "ssdef.h"      /* System Service constants - SS$_NORMAL etc.*/
#include "prvdef.h"     /* Process privilege defs and constants */
#include "jpidef.h"     /* GETJPI definitions and constants */
#include "Xlib.h"       /* True/False definitions */
/*-------------*
 *  Constants  *
 *-------------*/

#define DISABLE False           /* SYS$SETPRV() constant to disable privs */
#define ENABLE  True            /* SYS$SETPRV() constant to enable privs */
                                                          
/*---------------------*
 *  Type Definitions   *
 *---------------------*/

/*  Quadword for privilege mask */
typedef long PRIVILEGE_MASK[2];

typedef struct _getjpi_itmlst 
{
    short   w_buff_len;
    short   w_item_code;
    long    l_buff_addr;
    long    aw_bytes_ret;
    long    l_terminate;
} GETJPI_ITMLST;


change_privs(char mode, char num_privs, ...)
{

    char *mode_str;
    char *enabl = "Enabling";
    char *disab = "Disabling";
    va_list ap;
    int new_priv;
    int l_status;
    int i;
    PRIVILEGE_MASK new_priv_mask = {0,0};
    PRIVILEGE_MASK active_privs= {0,0};

    if (mode == ENABLE) 
    {
        mode_str = enabl;
        get_priv_mask(JPI$_CURPRIV,&active_privs);
    }
    else 
    {
        mode_str = disab;
        get_priv_mask(JPI$_PROCPRIV,&active_privs);
    }

    /*  
     *  Initialize the ap pointer to
     *  The start of the variable part
     *  of the argument list.
     */
    va_start(ap,num_privs);

    /*  Change Privs according to given mode */
    for (i=0; i<num_privs; i++) 
    {
        /*  
         *  Get a priv argument.
         */
        new_priv = va_arg(ap,int);

        /*
         *  If we are setting privs, active_privs[] is the
         *  current active privs.  We don't set an active priv.
         *
         *  If we are disabling privs, active_privs[] is the
         *  permanent process privs so we make sure we don't
         *  disable a priv the user originally posessed.
         */
            continue;       /* Skip this one and Do next priv argument */

        /*
         *  OR the value into the new_priv_mask.
         */
        new_priv_mask[0] |= new_priv;
        l_status = SYS$SETPRV (mode,&new_priv_mask,0,0);
        if (l_status != SS$_NORMAL) 
        {
            fprintf (stderr,
                     "$setprv returned %x when %s priv = %x in SECURITY code!",
                     l_status,
                     mode_str,
                     new_priv);
            exit(0);
        }
        new_priv_mask[0] = 0;
    }

    /*  End of the variable arguments */
    va_end(ap);

    return (SS$_NORMAL);
}

get_priv_mask(int mask_type, PRIVILEGE_MASK *q_priv_mask)
{
    int l_status;
    short w_ret_len = 0;
    GETJPI_ITMLST r_itmlst = {8,0,q_priv_mask,&w_ret_len,0};

    switch(mask_type) 
    {

        case JPI$_CURPRIV :
            r_itmlst.w_item_code = JPI$_CURPRIV;
            break;

        case JPI$_IMAGPRIV :
            r_itmlst.w_item_code = JPI$_IMAGPRIV;
            break;

        case JPI$_PROCPRIV :
            r_itmlst.w_item_code = JPI$_PROCPRIV;
            break;

        case JPI$_AUTHPRIV :
            r_itmlst.w_item_code = JPI$_AUTHPRIV;
            break;

        default:
            break;
    }

    l_status = SYS$GETJPI(0,0,0,&r_itmlst,0,0,0);
    if (l_status != SS$_NORMAL) 
    {
        fprintf(stderr,"GETJPI Failed with status = %x in SECURITY code!\n",
            l_status);
        exit(0);
    }
    return(SS$_NORMAL);
}
#endif /* VMS */
