
#ifndef __XTRAPDDMI__
#define __XTRAPDDMI__ /* "@(#)xtrapddmi.h	1.15 - 90/11/07  " */

/*****************************************************************************
Copyright 1987, 1988, 1989, 1990 by Digital Equipment Corporation, Maynard, MA

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

*****************************************************************************/
/*
 *
 *  CONTRIBUTORS:
 *
 *      Dick Annicchiarico
 *      Robert Chesler
 *      Dan Coutu
 *      Gene Durso
 *      Marc Evans
 *      Alan Jamison
 *      Mark Henry
 *      Ken Miller
 *
 *  DESCRIPTION:
 *      This header file is used by the XTrap server extension only
 *      (not used by clients or the XTrap Toolkit).  Information
 *      contained herein should *not* be visible to clients (xtrapdi.h
 *      is used for this).  The name is historical.
 */
#include "X.h"
#include "bits.h"
#include "dix.h"

#ifndef MIN
# define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif
#ifndef MAX
# define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

#define XETrapNumEvents    1L  /* constants used for AddExtension */

/* Other constants used within the extension code */
#define XETrapMinRepSize        32L        /* Minimum number of longs */

#ifdef VMS
typedef struct
{
    short    status;
    short    msg_len;
    long     unused;
} IOSB;                       /* I/O status block */

#ifdef DNETCONN      /* for non-transparent DECnet I/O */
#define MAX_NCB     110L
typedef struct
{
    short int msg;
    short int unit;
    char name_info[MAX_NCB];
} MSG_BLOCK;

typedef struct
{
    IOSB      iosb;
    MSG_BLOCK msg_block;
    short     chan;
} NETCMD;
#endif /* DNETCONN */
#endif /* VMS */

typedef union                    /* I/O thingy to ID what to access */
    {
       int       fd;             /* file descriptor */
       short     chan;           /* I/O channel */
    } XETrapID; 
typedef struct /* environment specific structure to support I/O transport */
{   /* Note: "streams" definition is *not* used in extension */
    int_function io;             /* pointer to XTrap I/O transport rtn */
    int_function close;          /* pointer to XTrap close transport rtn */
    XETrapID     id;             /* I/O object identifier (union) */
    CARD8        transport;      /* One of the known families to extension */
#ifdef vms
    int          mode;           /* Access Mode bit mask */
    IOSB         iosb;           /* I/O status block */
#endif /* VMS */
} XETrapIO;

/* This structure will be globally declared to provide storage to hold
 * the various extension wide configuration information.  Allocated on
 * a per-client basis.
 */
typedef struct
{
#ifndef _XINPUT
     /* Read buffer for receiving X events/req's. Must be quadword-aligned */
    XETrapInputReq       read_buff;
    XETrapIO             read;   /* Extension input transport struct */
#endif
    ClientPtr            client;  /* Multi-client support and error handling */
    XETrapIO             write;  /* Extension output transport struct */
    xXTrapGetCurReply    cur;    /* Struct of Miscellaneous state info */
    xXTrapGetStatsReply  *stats; /* Pointer to stat's, malloc'd if requested */
#if defined VMS && defined DNETCONN
    NETCMD               netcmd; /* VMS DECnet-specific fields */
#endif
    CARD32              last_input_time; /* last timestamp from input event */
} XETrapEnv;

#define XETrapSetHeaderEvent(phdr)      ((phdr)->type = 0x1L)
#define XETrapSetHeaderRequest(phdr)    ((phdr)->type = 0x2L)
#define XETrapSetHeaderSpecial(phdr)    ((phdr)->type = 0x3L)
#define XETrapSetHeaderCursor(phdr)     ((phdr)->type = 0x4L)

#ifndef X11R4
/* Swapping macros not visible in X11 Release 3 */
#define LengthRestB(stuff) \
    (((unsigned long)stuff->length << 2) - sizeof(*stuff))

#define LengthRestS(stuff) \
    (((unsigned long)stuff->length << 1) - (sizeof(*stuff) >> 1))

#define LengthRestL(stuff) \
    ((unsigned long)stuff->length - (sizeof(*stuff) >> 2))

#define SwapRestS(stuff) \
    SwapShorts((short *)(stuff + 1), LengthRestS(stuff))

#define SwapRestL(stuff) \
    SwapLongs((long *)(stuff + 1), LengthRestL(stuff))

/* copy long from src to dst byteswapping on the way */
#define cpswapl(Src, Dest) \
    { \
      char *pSrc = (char *)&(Src); \
      char *pDest = (char *)&(Dest); \
      pDest[0] = pSrc[3]; \
      pDest[1] = pSrc[2]; \
      pDest[2] = pSrc[1]; \
      pDest[3] = pSrc[0]; \
    }

/* copy short from src to dst byteswapping on the way */
#define cpswaps(Src, Dest)\
    { \
      char *pSrc = ((char *)&(Src)); \
      char *pDest = ((char *)&(Dest)); \
      pDest[1] = pSrc[0]; \
      pDest[0] = pSrc[1]; \
    }
#endif  /* X11R4 */

#ifndef vaxc
#define globaldef 
#define globalref extern
#endif
/* Current Release, Version, and Revision of the XTrap Extension */
#define XETrapRelease      3L
#define XETrapVersion      1L
#ifndef XETrapRevision        /* Changed from the Makefile by users */
# define XETrapRevision    0L
#endif /* XETrapRevision */

/* Extension platform identifier (conditionally defined) */
#ifdef ultrix
# define XETrapPlatform    PF_DECUltrix
#endif
#ifdef vms
# define XETrapPlatform    PF_DECVMS
#endif
#ifdef VT1000
# define XETrapPlatform    PF_DECVT1000
#endif
#ifdef VXT
# define XETrapPlatform    PF_DECXTerm
#endif
#ifdef PC
# define XETrapPlatform    PF_IBMAT
#endif
#ifdef sun
# define XETrapPlatform    PF_SunSparc
#endif
#ifndef XETrapPlatform
# define XETrapPlatform    PF_Other
#endif  /* XETrapPlatform */

#endif /* __XTRAPDDMI__ */
