/*****************************************************************************
Copyright 1987, 1988, 1989, 1990 by Digital Equipment Corporation, Maynard, MA

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

*****************************************************************************/
/*
 *  ABSTRACT:
 *
 *      This module is the VMS-specific module for extension I/O routines.
 *      It is called by the server and by clients using the extension.
 *      This is shared code and is subject to change only by team approval.
 *
 *  CONTRIBUTORS:
 *
 *      Dick Annicchiarico
 *      Robert Chesler
 *      Dan Coutu
 *      Gene Durso
 *      Marc Evans
 *      Alan Jamison
 *      Mark Henry
 *      Ken Miller
 *
 */
#ifndef lint
static char SCCSID[] = "@(#)xtrapddvms.c	1.18 - 90/09/18  ";
#endif

#ifdef vms
/*-----------------*
 *  Include Files  *
 *-----------------*/
#include "stdio.h"
#include "file.h"
#include "time.h"
#include "errno.h"
#include "types.h"

#define NEED_REPLIES
#define NEED_EVENTS
#include "X.h"                  /* General X11 definitions */
#include "Xproto.h"             /* X11 protocol definitions */
#include "input.h"		/* Server DevicePtr definitions */
#include "dixstruct.h"		/* Server ClientRec definitions */
#include "extnsionst.h"		/* Server ExtensionEntry definitions */
#include "malloc.h"             /* Memory allocation macros (Xalloc) */

#include "xtrapdi.h"            /* Device Independent XTrap Header File */
#include "xtrapddmi.h"          /* Device Dependent, Machine Indpendent */
#include "xtrapproto.h"         /* Function Prototypes for XTrap */
#include "bits.h"               /* For Bit Macros */
#include "scrnintstr.h"         /* screenInfo structure */
#include "iodef.h"              /* For VMS QIO definitions */
#include "stsdef.h"             /* For VMS Status identifiers */
#include "ssdef.h"              /* For VMS Status identifiers */
#include "msgdef.h"             /* For VMS Message identifiers */
#include "descrip.h"            /* For VMS descriptors */
#include "psldef.h"             /* Process Status Longword Definitions */

#define $DESCRIPTOR_R(dsc_name, dsc_size, dsc_buff)\
            struct dsc$descriptor_s _align (LONGWORD) dsc_name = \
            { dsc_size, DSC$K_DTYPE_T, DSC$K_CLASS_S, dsc_buff }

/*----------------------------*
 *  Global Data Declarations  *
 *----------------------------*/

/* This carries all of the information XTrap uses for internal information */
extern int_function XETrapIoConf[];   /* IO configuration functions */
extern xXTrapGetAvailReply XETrap_avail;
extern int XETrapErrorBase;

#ifdef DEBUG
extern FILE *dfd;
#endif /* DEBUG */

/*---------------------*
 *  Forward References *
 *---------------------*/
#ifndef _XINPUT
static void read_ast();
#endif
static void write_ast();


#ifndef _XINPUT
/*
 *  DESCRIPTION:
 *
 *  This routine processes the simulation of some input event.  
 *  VMS implementation below uses SimInput Utility to
 *  place events directly on the server's input queue.  It is
 *  intended to utilize the X Input Extension in place of
 *  SimInput() once it becomes more readily available.
 *
 */
int 
#ifdef __STDC__
XETrapSimulateXEvent( xXTrapInputReq *req, ClientPtr client)
#else
XETrapSimulateXEvent(req, client)
    xXTrapInputReq *req;
    ClientPtr client;
#endif
{

    register int status = Success;

    /* Check to make sure screen number is valid for this server */
    if (screenInfo.numScreens < req->input.screen)
    {   /* Trying to play bogus events to this WS! */
#ifdef VERBOSE
        ErrorF("%s:  Trying to send events to screen %d!\n", XTrapExtName,
            req->input.screen);
#endif
        status = XETrapErrorBase + BadScreen;
    }
    else if (!(SimInputNew(req->input.type, req->input.detail, 
        req->input.x, req->input.y, req->input.screen)))
    {
#ifdef VERBOSE
        ErrorF("%s:  SimInput Error!\n", XTrapExtName);
#endif
        status = BadImplementation;
    }
    return(status);
}
#endif
    
#ifndef _XINPUT
/*
 *  DESCRIPTION:
 *
 *      This function performs the transport specific functions for reading
 *      data from an XTrap client via a generic File VMS I/O channel.
 *      The read completion ast performs the actual SimInput call as well
 *      as posting another read
 *
 */
#ifdef __STDC__
int XETrapReadChan( XETrapEnv *penv)
#else
int XETrapReadChan(penv)
    XETrapEnv *penv;
#endif
{

    if (penv->read.id.chan)
    {
        int vms_status;
        if ((vms_status = sys$qio(0L,penv->read.id.chan,penv->read.mode,
            &penv->read.iosb, read_ast, penv, &penv->read_buff, 
            sizeof(penv->read_buff), 0L, 0L,0L,0L)) == SS$_NORMAL)
        {
            return(sizeof(penv->read_buff));  /* actually, "expected" */
        }
        ErrorF("%s:  VMS Error %%x%x in XETrapReadChan!\n",
            XTrapExtName, vms_status);
    }
    return 0L;  /* No data read, cause no I/O identifier or error on qio */
}
#endif

/*
 *  DESCRIPTION:
 *
 *      This function performs the transport specific functions for writing
 *      data back to an XTrap client via a generic File VMS I/O channel.
 *
 */
#ifdef __STDC__
int XETrapWriteChan( XETrapEnv *penv, BYTE *data, CARD16 size)
#else
int XETrapWriteChan(penv, data, size)
    XETrapEnv *penv;
    BYTE      *data;
    CARD16    size;
#endif
{
#ifdef DEBUG
    int  i;
    unsigned char ch;
    char *ptr;
    fprintf(dfd,"QIO: (%d) ", size);
    ptr = data;
    for (i=0; i<size; i++)
    {
        ch = *ptr++;
        fprintf(dfd, "%02x ", ch);
    }
    fprintf(dfd, "\n");
#endif /* DEBUG */
 
    if (penv->write.id.chan)
    {
        int vms_status;
        if ((vms_status = sys$qio(0L,penv->write.id.chan,penv->write.mode,
            &penv->write.iosb, write_ast,&penv->write.iosb, data, size, 
            0L, 0L,0L,0L)) == SS$_NORMAL)
        {
            return(size);
        }
        ErrorF("%s:  VMS Error %%x%x in XETrapWriteChan!\n",
            XTrapExtName, vms_status);
    }
    return 0L;  /* No data sent, cause no I/O identifier or error on qio */
}

/*
 *  DESCRIPTION:
 *
 *      This function performs the transport specific functions for closing
 *      down a VMS I/O channel.
 *
 */
#ifdef __STDC__
int XETrapCloseChan( XETrapID id)
#else
int XETrapCloseChan(id)
    XETrapID id;
#endif
{

    int status;

    if (status = sys$dassgn(id.chan) != SS$_NORMAL)
    {
        ErrorF("%s:  VMS Error %%x%x in XETrapCloseChan\n", 
                XTrapExtName,status);
        return(EOF);
    }
    return(0L);
}

#ifndef _XINPUT
/*
 *  DESCRIPTION:
 *
 *      This function performs the VMS read completion at AST level.  
 *      Status checking, input simulation, and posting another read
 *      is required.
 *
 */
#ifdef __STDC__
static void read_ast( XETrapEnv *penv)
#else
static void read_ast(penv)
    XETrapEnv *penv;
#endif
{

    XETrapInputReq *inp = &penv->read_buff;
    if ((penv->read.iosb.status == SS$_LINKDISCON) ||
        (penv->read.iosb.status == SS$_ABORT) ||
        (penv->read.iosb.status == SS$_LINKABORT))
    {   /* Client disconnected */
        return;
    }
    if (penv->read.iosb.status != STS$M_SUCCESS)
    {   /* Read wasn't successful */
        ErrorF("%s:  VMS Error %%x%x in XETrapReadAST!\n",
            XTrapExtName, penv->read.iosb.status);
    }
    else
    {   /* Generate the requested X Event */
        if (!SimInputNew(inp->type, inp->detail, inp->x, inp->y, inp->screen))
        {
            ErrorF("%s:  SimInput Error in XETrapReadAST!\n",
                XTrapExtName);
        }
    }
    if ((penv->read.io) && (penv->read.iosb.status != SS$_LINKABORT))
    {   /* Now post another read */
        if ((*penv->read.io)(penv) == 0L)
        {
            ErrorF("%s:  Error Reading Input!\n", XTrapExtName);
        }
    }
    return;
}
#endif

/*
 *  DESCRIPTION:
 *
 *      This function performs the VMS write completion at AST level.  
 *      Basically, all that's required is status checking.
 *
 */
#ifdef __STDC__
static void write_ast( IOSB *iosb)
#else
static void write_ast(iosb)
    IOSB *iosb;
#endif
{

    if (iosb->status != STS$M_SUCCESS)
    {
        ErrorF("%s:  VMS Error %%x%x in XETrapWriteAST!\n",
            XTrapExtName,iosb->status);
    }
    return;
}

#ifdef DNETCONN
static void issue_netcmd_read();

#ifdef __STDC__
static void process_netcmd( NETCMD *netcmd)
#else
static void process_netcmd(netcmd)
    NETCMD *netcmd;
#endif
{

    int status = True;

    switch(netcmd->iosb.status)
    {   /* the following are "normal" I/O return
         * values received in course of connection
         */
        case SS$_NORMAL:
            break;
        case SS$_CANCEL:
        case SS$_ABORT:
            if (status = sys$dassgn(netcmd->chan) != SS$_NORMAL)
            {
                ErrorF("%s:  VMS Error %%x%x in XETrapProcessNetCmd\n",
                        XTrapExtName,status);
            }
            status = False;
            break;
        default:
            ErrorF("%s:  VMS I/O Error %%x%x in process_netcmd!\n",
                XTrapExtName, netcmd->iosb.status);
            break;
    }
    switch(netcmd->msg_block.msg)
    {   /* the following are "normal" messages 
         * received in course of connection
         */
        case MSG$_CONNECT:
        case MSG$_CONFIRM:
        case MSG$_INTMSG:
            break;
        case MSG$_ABORT:
            if (status = sys$dassgn(netcmd->chan) != SS$_NORMAL)
            {
                ErrorF("%s:  VMS Error %%x%x in XETrapProcessNetCmd\n",
                        XTrapExtName,status);
            }
            status = False;
            break;
        default:
            ErrorF("%s:  VMS Msg Error %d in process_netcmd!\n",
                XTrapExtName, netcmd->msg_block.msg);
            break;
    }
    if (status == True)
    {
        issue_netcmd_read(netcmd);
    }
}
#endif
#ifdef DNETCONN
#ifdef __STDC__
static void issue_netcmd_read( NETCMD *netcmd)
#else
static void issue_netcmd_read(netcmd)
    NETCMD *netcmd;
#endif
{

    int vms_status;

    if ((vms_status = sys$qio(0L, netcmd->chan, IO$_READVBLK,
        &netcmd->iosb, process_netcmd, netcmd, &netcmd->msg_block, 
        sizeof(MSG_BLOCK), 0L, 0L, 0L, 0L)) != SS$_NORMAL)
    {
        ErrorF("%s:  VMS Read Error %%x%x in issue_netcmd_read!\n",
            XTrapExtName, vms_status);
    }
}
#endif
#ifdef DNETCONN

/*
 *  DESCRIPTION:
 *
 *      This function performs the VMS DECnet initialization. Returns
 *      a channel or -1 if failure.
 *
 */
#ifdef __STDC__
short XETrapInitDECnetVMS( char *object, XETrapEnv *penv)
#else
short XETrapInitDECnetVMS(object, penv)
    char      *object;
    XETrapEnv *penv;
#endif
{

    int vms_status;
    short chan;
    IOSB iosb;
    $DESCRIPTOR_R(dsc_object, strlen(object), object);
    int max_msg = 128L;
    int buf_quo = 128L;
    static $DESCRIPTOR (net_device, "_NET:");

    if ((vms_status = lib$asn_wth_mbx(&net_device, &max_msg, &buf_quo, 
        &chan,&(penv->netcmd.chan))) != SS$_NORMAL)
    {
        chan = -1L;
        ErrorF("%s:  VMS Assignment Error %%x%x in XETrapInitDECnet!\n",
            XTrapExtName, vms_status);
    }
    if (((vms_status = sys$qiow(0L, chan, IO$_ACCESS, &iosb, 0L, 0L, 0L,
            &dsc_object, 0L, 0L, 0L, 0L)) != SS$_NORMAL) ||
            (iosb.status != SS$_NORMAL))
    {
        chan = -1L;
        ErrorF("%s:  VMS Network Error %%x%x, %%x%x in XETrapInitDECnet!\n",
            XTrapExtName, vms_status, iosb.status);
    }
    if (chan != -1L)
    {
        issue_netcmd_read(&(penv->netcmd));
    }

    return(chan);
}
#endif

#ifdef DNETCONN
/*
 *  DESCRIPTION:
 *      This function performs the transport specific functions for 
 *      closing the DECnet connection.  This is different from CloseChan
 *      in that a QIO to de-access from the net is req'd.  Following
 *      the de-allocation, it just calls XETrapCloseChan.
 *
 */
#ifdef __STDC__
int XETrapCloseDECnetVMS( XETrapID id)
#else
int XETrapCloseDECnetVMS(id)
    XETrapID id;
#endif
{

    int vms_status;
    if ((vms_status = sys$qiow(0L,id.chan,IO$_DEACCESS|IO$M_SYNCH,
        0L,0L,0L,0L,0L,0L,0L,0L,0L)) != SS$_NORMAL)
    {
        ErrorF("%s:  VMS Error %%x%x in XETrapCloseDECnet!\n",
            XTrapExtName, vms_status);
    }

    return(XETrapCloseChan(id));
}
#endif

/*
 *  DESCRIPTION:
 *
 *      This function performs the VMS Mailbox initialization.
 *
 */
#ifdef __STDC__
short XETrapInitMbxVMS( char *mbx)
#else
short XETrapInitMbxVMS(mbx)
    char *mbx;
#endif
{

    short chan;
    int vms_status;
    $DESCRIPTOR_R(dsc_mbx, strlen(mbx), mbx);

    if ((vms_status = sys$assign(&dsc_mbx, &chan, PSL$C_USER, 0L))
        != SS$_NORMAL)
    {
        chan = -1L;
#ifdef VERBOSE
        ErrorF("%s:  VMS Error %%x%x in XETrapInitMbx!\n",
             XTrapExtName, vms_status);
#endif
    }
    return(chan);
}

/*
 *  DESCRIPTION:
 *
 *      This function performs the transport VMS specific operations to
 *      connect to the client application via mailbox
 */
#ifdef __STDC__
int XETrapConfigMbx( xXTrapConfigReq *req, XETrapEnv *penv)
#else
int XETrapConfigMbx(req,penv)
    xXTrapConfigReq *req;
    XETrapEnv *penv;
#endif
{

    int status = Success;
    short chan;
    char *out_mbx = req->config.address;
    char *inp_mbx = &req->config.address[strlen(req->config.address)+1L];

    if (BitIsTrue(penv->cur.data.config.flags.data,XETrapTransOut))
    {   /* We are writing */
        penv->write.io        = XETrapWriteChan;
        penv->write.close     = XETrapCloseChan;
        penv->write.id.chan   = XETrapInitMbxVMS(out_mbx);
        penv->write.mode      = IO$_WRITEVBLK | IO$M_NOW;
        penv->write.transport = FamilyMbx;
        if (penv->write.id.chan < 0L)
        {
            status = XETrapErrorBase + BadMailbox;
        }
    }
#ifndef _XINPUT
    if ((status == Success) && 
        (BitIsTrue(penv->cur.data.config.flags.data,XETrapTransIn)))
    {   /* We're reading */
        SimInit();          /* Initialize SimInput */
        penv->read.io        = XETrapReadChan;
        penv->read.close     = XETrapCloseChan;
        penv->read.id.chan   = XETrapInitMbxVMS(inp_mbx);
        penv->read.mode      = IO$_READVBLK;
        penv->read.transport = FamilyMbx;
        if (penv->read.id.chan < 0L)
        {
            status = XETrapErrorBase + BadMailbox;
        }
        /* Now post a read for incoming requests */
        else if ((*penv->read.io)(penv) == 0L)
        {
            status = XETrapErrorBase + BadIO;
        }
    }
#endif
    return(status);
}

/* Used by swapping code not visible from VMS (from main.c) */
void
#ifdef __STDC__
NotImplemented({)
#else
NotImplemented()
#endif
{

    FatalError("Not implemented");
}

int
#ifdef __STDC__
ProcBadRequest( ClientPtr client)
#else
ProcBadRequest(client)
    ClientPtr client;
#endif
{

    return (BadRequest);
}

#endif /* vms */
